/**********************************************************************
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 **********************************************************************/

#ifndef F_CPU
	#define F_CPU 12000000
#endif

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "i2clcd.h"

#define ADDR_TASTMAT 0x42


//*********************************************************
unsigned char st=2;           /* Spielstufe (Level) */

/* Attention: union TIMER assumes little endian CPU or compiler!           */
typedef union {
	long w;
	unsigned char c[4];
} TIMER;

TIMER timer;                    // wird alle 4.096ms erhoeht

#define BLINKOFF  0
#define BLINKALL  1
#define BLINKDOT  2
#define BLINKRUN  4

#define FN  '9'
#define CL  ':'
#define GO  ';'

/* better readability of AVR Program memory arrays */
#define o(ndx)	(signed char)pgm_read_byte(o+(ndx))
#define w(ndx)	(signed char)pgm_read_word(w+(ndx))

volatile unsigned char keycode; // 0 fuer keine Taste, 1..11 fuer Taste

//*********************************************************


/* -------------------------------------------------------------------------- */
// Aufruf: 12MHz / 1024 / 65488 = 244Hz = 4.096ms
#define PRELOAD		65488
ISR(TIMER1_OVF_vect) {
	TCNT1     = PRELOAD;
	++timer.w;
}


/* -------------------------------------------------------------------------- */
void init(void) {
	// ToDo --> Timer initialisieren 244Hz...
	TCNT1   = PRELOAD;						// Preloadwert setzen
	TCCR1B |= (1 << CS02) | (1 << CS00);	// Prescaler Timer1 auf 1024 stellen
	TIMSK  |= (1 << TOIE1);					// Timer 1 Overflow Interrupt enable
	sei(); 									// Interrupts einschalten

}

/* -------------------------------------------------------------------------- */
void long_delay(uint16_t ms) {				// lange Pause
    for (; ms>0; ms--) _delay_ms(1);
}

/* -------------------------------------------------------------------------- */
uint8_t keyboard_write(uint8_t val) {
	if (i2c_start(ADDR_TASTMAT+I2C_WRITE)) return 1;
	i2c_write(val);
	i2c_stop();
	return 0;
}

/* -------------------------------------------------------------------------- */
uint8_t keyboard_read(void) {
	uint8_t val;
	if (i2c_start(ADDR_TASTMAT+I2C_READ)) return 0;
	val = i2c_readNak();
	i2c_stop();
	return val;
}


/* -------------------------------------------------------------------------- */
uint8_t keyboard_init(void) {
	// Zeilen auf High, Spalten auf Low...
	// wird eine Taste gedrueckt, aendert sich der Pegel an einer der
	// Spalten und es wird ein Interrupt ausgeloest
	if (keyboard_write(0x0F)) return 1;
	return 0;
}

/* -------------------------------------------------------------------------- */
uint8_t keyboard_scan(void) {
	uint8_t row, col, key;
	row = keyboard_read();				
	if (row < 0x0F) {					// einer der Zeilen ist durchgeschalten
		keyboard_write(0xF0);			// Spalten auslesen
		col = keyboard_read();			// ...
		key = col+row;					// Keycode berechnen
		keyboard_init();				// Ausgangszustand (Zeilen ueberwachen)
		long_delay(90);					// eine Weile warten wg. Tastenprellen
		row = keyboard_read();			// und das Gleiche nochmal...
		if (row < 0x0F) {				// ... wegen Tastenprellen
			keyboard_write(0xF0);		// ...
			col = keyboard_read();
			keyboard_init();
		} else {
			return 0;	
		}
		if (key == (col+row)) {			// nochmal die gleiche Taste? 
			switch (key)				// Tastencode in was "lesbares" umsetzen
				{
					case 238 : return 1  + '0';	
					case 237 : return 2  + '0';	
					case 235 : return 3  + '0';	
					case 231 : return 4  + '0';	
					case 222 : return 5  + '0';	
					case 221 : return 6  + '0';	
					case 219 : return 7  + '0';	
					case 215 : return 8  + '0';	
					case 190 : return 9  + '0';	
					case 189 : return 10 + '0';	
					case 187 : return 11 + '0';	
					case 183 : return 12 + '0';	
					case 126 : return 13 + '0';	
					case 125 : return 14 + '0';	
					case 123 : return 15 + '0';	
					case 119 : return 16 + '0';	
				}
			return key;					
		} else {						
			return 0;					
		}		
	}
	return 0;	
}

/* -------------------------------------------------------------------------- */
uint8_t getkey(void) {
	uint8_t key; 
	do 
		key = keyboard_scan();
	while (!key);
	return key;
		
}

/* -------------------------------------------------------------------------- */
#define COLUMN 1
#define LINE   2
void putled(uint8_t offset, uint8_t c) {
	if ((c>='a' && c <= 'z') ||
	    (c>='A' && c <= 'Z') ||
	    (c>='0' && c <= '9'))	{
			lcd_putcharlr(LINE, COLUMN+offset, c);
	} else {
			lcd_putcharlr(LINE, COLUMN+offset, c);
	}
}

/* -------------------------------------------------------------------------- */
void blink(uint8_t mode) {
	switch (mode) {
		case 0: lcd_printlc(1, 1, (unsigned char *)"Mein Zug... "); break;
		case 1: lcd_printlc(1, 1, (unsigned char *)"BLINKALL????"); break;
		case 2: lcd_printlc(1, 1, (unsigned char *)"BLINKDOT????"); break;
		case 4: lcd_printlc(1, 1, (unsigned char *)"Ich denke..."); break;
	}
}

/* -------------------------------------------------------------------------- */
void clean(void) {
  putled(0, ' ');
  putled(1, ' ');
  putled(2, ' ');
  putled(3, ' ');
}

void bold(void) {
	// erstmal dummy...
	// eigentlich soll hier lt. Orginal die Anzeige eingeschaltet 
	// werden
}


/***************************************************************************/
/*                               micro-Max,                                */
/* A chess program smaller than 2KB (of non-blank source), by H.G. Muller  */
/***************************************************************************/
/* version 4.8 (1953 characters) features:                                 */
/* - recursive negamax search                                              */
/* - all-capture MVV/LVA quiescence search                                 */
/* - (internal) iterative deepening                                        */
/* - best-move-first 'sorting'                                             */
/* - a hash table storing score and best move                              */
/* - futility pruning                                                      */
/* - king safety through magnetic, frozen king in middle-game              */
/* - R=2 null-move pruning                                                 */
/* - keep hash and repetition-draw detection                               */
/* - better defense against passers through gradual promotion              */
/* - extend check evasions in inner nodes                                  */
/* - reduction of all non-Pawn, non-capture moves except hash move (LMR)   */
/* - full FIDE rules (expt under-promotion) and move-legality checking     */

#define M 0x88                              /* Unused bits in valid square */  
#define S 128                               /* Sign bit of char            */
#define I 8000                              /* Infinity score              */
#define U 18                                /* D() Stack array size        */

struct {
short q,l,e;          /* Args: (q,l)=window, e=current eval. score         */
short m,v,            /* m=value of best move so far, v=current evaluation */
 V,P;
unsigned char E,z,n;  /* Args: E=e.p. sqr.; z=level 1 flag; n=depth        */          
signed char r;        /* step vector current ray                           */
unsigned char j,      /* j=loop over directions (rays) counter             */ 
 B,d,                 /* B=board scan start, d=iterative deepening counter */ 
 h,C,                 /* h=new ply depth?, C=ply depth?                    */
 u,p,                 /* u=moving piece, p=moving piece type               */
 x,y,                 /* x=origin square, y=target square of current move  */
 F,                   /* F=e.p., castling skipped square                   */
 G,                   /* G=castling R origin (corner) square               */
 H,t,                 /* H=capture square, t=piece on capture square       */
 X,Y,                 /* X=origin, Y=target square of best move so far     */
 a;                   /* D() return address state                          */
} _, A[U],*J=A+U;     /* _=working set, A=stack array, J=stack pointer     */

short Q,                                    /* pass updated eval. score    */
K,                                          /* input move, origin square   */
i,s,                                        /* temp. evaluation term       */
Dq,Dl,De,                         /* D() arguments */
DD;                               /* D() return value */

const signed char w[] PROGMEM = {
  0,2,2,7,-1,8,12,23};                         /* relative piece values    */

const signed char o[] PROGMEM = {
  -16,-15,-17,0,1,16,0,1,16,15,17,0,14,18,31,33,0,    /* step-vector lists */
  7,-1,11,6,8,3,6,                             /* 1st dir. in o[] per piece*/
  6,3,5,7,4,5,3,6};                            /* initial piece setup      */
  
unsigned char L,                               /* input move, target square*/
b[129],                                        /* board: half of 16x8+dummy*/
c[5],                                          /* input move ASCII buffer  */     
k,                                          /* moving side 8=white 16=black*/
O,                                          /* pass e.p. flag at game level*/
R,                                          /* captured non pawn material  */
DE,Dz,Dn,                                      /* D() arguments            */
Da,                                            /* D() state                */
W,                                           /* @ temporary                */
hv;                                            /* zeige Hauptvariante      */

void D(void);






/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
int main(void) {
	//char s[10];
	//uint8_t key;

	i2c_init();
	lcd_init();
	keyboard_init();

	init();		// Chessinit...


	lcd_command(LCD_CLEAR);
	lcd_backlight(ON);
	lcd_printlc(1,1,(unsigned char *)"-->BLIT-Board<--");
	lcd_printlc(2,1,(unsigned char *)"Uwe Berger; 2008");
	long_delay(1000);
	lcd_command(LCD_CLEAR);

	while (1) {								

// ab hier aus der Quelle entnommen...
A:blink(BLINKOFF);
  putled(0, 'S');putled(1, 'H');putled(2, 'A');putled(3, 'H');
  k=16;O=Q=R=0;
  for(W=0;W<sizeof b;++W)b[W]=0;
  W=8;while(W--) 
  {b[W]=(b[W+112]=o(W+24)+8)+8;b[W+16]=18;b[W+96]=9; /* initial board setup*/
   L=8;while(L--)
    b[16*L+W+8]=(W-4)*(W-4)+(L+L-7)*(L+L-7)/4;       /* center-pts table   */
  }                                                  /*(in unused half b[])*/

  do{                                                /* play loop          */
   W = 4;
   for(;;) {
    switch (c[W] = getkey()) {
    case FN:
      putled(0, 'F');putled(1, 'N');putled(2, ' ');putled(3, ' ');
      switch (L = getkey()) {
      case '1': goto A; break;      // Neues Spiel (new game)
      case '2':                     // Spielstufe (set level)
        for(;;) {
          putled(0, 'S');putled(1, 'T');putled(3, st+'1');
          if (GO == (L = getkey())) break;
          st=L-'1';
        }
        break;
      case '3':                     // Hauptvariante (show PV)
        for(;;) {
          putled(0, 'H');putled(1, 'V');putled(3, hv+'0');
          if (GO == (L = getkey())) break;
          hv=L&1;
        }
        break;
      }
      // no break
    case CL: W=5; break;
    case GO: goto B; break;
    }
    if (!(W&1)) {          // wahr bei W ist gerade (0, 2, 4)
      c[W] += 'a' - '1';
    }
    switch (W) {
    default:
	  putled(W, c[W]);
	  ++W;
      break;
    case 4:                 // getkey() anzeigen
	  clean();
	  c[0]=c[4];
	  putled(0, c[0]);
	  W=1;
      break;
    case 5:                 // getkey() nicht anzeigen
	  clean();
      W=0;
      break;
    }
   }   
B: 
   if(*c-GO)K=*c-16*c[1]+799,L=c[2]-16*c[3]+799;     /* parse entered move */
   else {
     K=I;
     cli();                   // Interrupts sperren
     timer.w= -128<<st;                              /* set time control   */
     sei();                   // Interrupts erlauben
     clean();
     if (hv) { bold(); blink(BLINKALL);}
     else blink(BLINKRUN);
   }
   Dq=-I;Dl=I;De=Q;DE=O;Dz=1;Dn=3;               /* store arguments of D() */
   Da=0;                                         /* state */
   D();           
   blink(BLINKOFF);
   if(*c-GO) {          // Zugeingabe vorhanden
    if(I==DD) {         // Zug ist gueltig
     bold();
    } else {            // .. ungueltig
     putled(0, 'Z');putled(1, 'U');putled(2, 'G');putled(3, ' ');
    } 	
    *c=GO;
   } else {             // Computerzug
    putled(0,'a'+(K&7));putled(1,'8'-(K>>4));  
    putled(2,'a'+(L&7));putled(3,'8'-(L>>4&7)); 
   }   
  }while(DD>-I+1);
  putled(0, 'M');putled(1, 'A');
  putled(2, 'T');putled(3, 'T');
  getkey();

	}
	return 0;
}



/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/* better readability of working struct variables */
#define q _.q
#define l _.l
#define e _.e
#define E _.E
#define z _.z
#define n _.n
#define m _.m
#define v _.v
#define V _.V
#define P _.P
#define r _.r
#define j _.j
#define B _.B
#define d _.d
#define h _.h
#define C _.C
#define u _.u
#define p _.p
#define x _.x
#define y _.y
#define F _.F
#define G _.G
#define H _.H
#define t _.t
#define X _.X
#define Y _.Y
#define a _.a

void D()                                       /* iterative Negamax search */
{                       
D:if (--J<A) {               /* stack pointer decrement and underrun check */
  ++J;DD=-l;goto R;                                    /* simulated return */
 } 
 q=Dq;l=Dl;e=De;E=DE;z=Dz;n=Dn;                          /* load arguments */
 a=Da;                                         /* load return address state*/

 --q;                                          /* adj. window: delay bonus */
 k^=24;                                        /* change sides             */
 d=X=Y=0;                                      /* start iter. from scratch */
 while(d++<n||d<3||                            /* iterative deepening loop */
   z&K==I&&(timer.w<0&d<98||                   /* root: deepen upto time   */
   (K=X,L=Y&~M,d=3)))                          /* time's up: go do best    */
 {x=B=X;                                       /* start scan at prev. best */
  h=Y&S;                                       /* request try noncastl. 1st*/
  if(d<3)P=I;else 
  {*J=_;Dq=-l;Dl=1-l;De=-e;DE=S;Dz=0;Dn=d-3;   /* save locals, arguments   */
   Da=0;goto D;                                /* Search null move         */
R0:_=*J;P=DD;                                  /* load locals, return value*/
  }
  m=-P<l|R>35?d>2?-I:e:-P;                     /* Prune or stand-pat       */
#ifdef __unix__
  ++timer.w;                                   /* node count (for timing)  */
#endif  
  do{u=b[x];                                   /* scan board looking for   */
   if(u&k)                                     /*  own piece (inefficient!)*/
   {r=p=u&7;                                   /* p = piece type (set r>0) */
    j=o(p+16);                                 /* first step vector f.piece*/
    while(r=p>2&r<0?-r:-o(++j))                /* loop over directions o[] */
    {A:                                        /* resume normal after best */
     y=x;F=G=S;                                /* (x,y)=move, (F,G)=castl.R*/
     do{                                       /* y traverses ray, or:     */
      H=y=h?Y^h:y+r;                           /* sneak in prev. best move */
      if(y&M)break;                            /* board edge hit           */
      m=E-S&b[E]&&y-E<2&E-y<2?I:m;             /* bad castling             */
      if(p<3&y==E)H^=16;                       /* shift capt.sqr. H if e.p.*/
      t=b[H];if(t&k|p<3&!(y-x&7)-!t)break;     /* capt. own, bad pawn mode */
      i=37*w(t&7)+(t&192);                     /* value of capt. piece t   */
      m=i<0?I:m;                               /* K capture                */
      if(m>=l&d>1)goto J;                      /* abort on fail high       */

      v=d-1?e:i-p;                             /* MVV/LVA scoring          */
      if(d-!t>1)                               /* remaining depth          */
      {v=p<6?b[x+8]-b[y+8]:0;                  /* center positional pts.   */
       b[G]=b[H]=b[x]=0;b[y]=u|32;             /* do move, set non-virgin  */
       if(!(G&M))b[F]=k+6,v+=50;               /* castling: put R & score  */
       v-=p-4|R>29?0:20;                       /* penalize mid-game K move */
       if(p<3)                                 /* pawns:                   */
       {v-=9*((x-2&M||b[x-2]-u)+               /* structure, undefended    */
              (x+2&M||b[x+2]-u)-1              /*        squares plus bias */
             +(b[x^16]==k+36))                 /* kling to non-virgin King */
             -(R>>2);                          /* end-game Pawn-push bonus */
        V=y+r+1&S?647-p:2*(u&y+16&32);         /* promotion or 6/7th bonus */
        b[y]+=V;i+=V;                          /* change piece, add score  */
       }
       v+=e+i;V=m>q?m:q;                       /* new eval and alpha       */
       C=d-1-(d>5&p>2&!t&!h);
       C=R>29|d<3|P-I?C:d;                     /* extend 1 ply if in check */
       do
        if(C>2|v>V)
        {*J=_;Dq=-l;Dl=-V;De=-v;DE=F;Dz=0;Dn=C; /* save locals, arguments  */
         Da=1;goto D;                          /* iterative eval. of reply */
R1:      _=*J;s=-DD;                           /* load locals, return value*/
        }else s=v;                             /* or fail low if futile    */
       while(s>q&++C<d);v=s;
       if(z&&K-I&&v+I&&x==K&y==L)              /* move pending & in root:  */
       {Q=-e-i;O=F;                            /*   exit if legal & found  */
        R+=i>>7;++J;DD=l;goto R;               /* captured non-P material  */
       }
       b[G]=k+6;b[F]=b[y]=0;b[x]=u;b[H]=t;     /* undo move,G can be dummy */
      }
      if(v>m)                                  /* new best, update max,best*/
       m=v,X=x,Y=y|S&F;                        /* mark double move with S  */
      if(h){h=0;goto A;}                       /* redo after doing old best*/
      if(x+r-y|u&32|                           /* not 1st step,moved before*/
         p>2&(p-4|j-7||                        /* no P & no lateral K move,*/
         b[G=x+3^r>>1&7]-k-6                   /* no virgin R in corner G, */
         ||b[G^1]|b[G^2])                      /* no 2 empty sq. next to R */
        )t+=p<5;                               /* fake capt. for nonsliding*/
      else F=y;                                /* enable e.p.              */
     }while(!t);                               /* if not capt. continue ray*/
  }}}while((x=x+9&~M)-B);                      /* next sqr. of board, wrap */
J:if(m>I-M|m<M-I)d=98;                         /* mate holds to any depth  */
  m=m+I|P==I?m:0;                              /* best loses K: (stale)mate*/
   if(z&hv&d>2)                                   
   {putled(0,'a'+(X&7));putled(1,'8'-(X>>4));  /* show Principal variation */
    putled(2,'a'+(Y&7));putled(3,'8'-(Y>>4&7)); 
   }
 }                                             /*    encoded in X S,8 bits */
 k^=24;                                        /* change sides back        */
 ++J;DD=m+=m<e;                                /* delayed-loss bonus       */
R:if (J!=A+U) switch(a){case 0:goto R0;case 1:goto R1;}
 else return;
}
